# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------------
#
# This script adopts 3-point method (Hasbargen, 2012) to calculate orientation of geologic surfaces.
# Inputs: point feature, where each of three points is located on the same geologic surface; raster DEM
# Outputs: point features (output3pointSHP - 3D dataset with results of calculations; outputCentroidSHP - 3D dataset with centroids of each triangle)
# Author: Andrzej Swiader, AGH UST Krakow
#
# ---------------------------------------------------------------------------

# Import requied modules
import os, arcpy, numpy, math

# Script arguments
input3pointSHP = arcpy.GetParameterAsText(0)
if input3pointSHP == '#' or not input3pointSHP:
    input3pointSHP = "D:\\dane\\doktorat\\doktorat_new.gdb\\geol\\geol_3point2D" # provide a default value if unspecified

inputDEM = arcpy.GetParameterAsText(1)
if inputDEM == '#' or not inputDEM:
    inputDEM = "D:\\dane\\doktorat\\LiDAR\\otryt_lidar.tif" # provide a default value if unspecified

output3pointSHP = arcpy.GetParameterAsText(2)
if output3pointSHP == '#' or not output3pointSHP:
    output3pointSHP = "D:\\dane\\doktorat\\doktorat_new.gdb\\geol\\geol_3point3D" # provide a default value if unspecified

outputCentroidSHP = arcpy.GetParameterAsText(3)
if outputCentroidSHP == '#' or not outputCentroidSHP:
    outputCentroidSHP = "D:\\dane\\doktorat\\doktorat_new.gdb\\geol\\geol_centroid3D" # provide a default value if unspecified

# Get the spatial reference from input dataset
spatialReference = arcpy.Describe(input3pointSHP).spatialReference

# Check number of objects (points, represented by rows in an attribute table) in an input dataset
nRows = int(arcpy.GetCount_management(input3pointSHP).getOutput(0))

# Set Geoprocessing environments
arcpy.env.outputZFlag = "Enabled"

# Process: Interpolate Shape
arcpy.InterpolateShape_3d(inputDEM, input3pointSHP, output3pointSHP, "", "1", "BILINEAR", "DENSIFY", "0")

# Process: Add Field surfID (Long integer) to bind each three subsequent points
arcpy.AddField_management(output3pointSHP, "surfID", "LONG", "", "", "", "", "NULLABLE", "NON_REQUIRED", "")

# Process: Add Geometry Attributes
arcpy.AddGeometryAttributes_management(output3pointSHP, "POINT_X_Y_Z_M", "", "", "")

# Add remaining necessarry fields
narray = numpy.array([],
                    numpy.dtype([('_ID', numpy.int),
                                 ('U1i', numpy.float),
                                 ('minusU2j', numpy.float),
                                 ('U3k', numpy.float),
                                 ('eastingFl', numpy.float),
                                 ('northingFl', numpy.float),
                                 ('strikeFl', numpy.float),
                                 ('dipFl', numpy.float),
                                 ('strike', numpy.int),
                                 ('dipDir', numpy.int),
                                 ('dip', numpy.int),
                                 ('easting', numpy.float),
                                 ('northing', numpy.float),
                                 ('elevation', numpy.float),
                                 ]))
 
arcpy.da.ExtendTable(output3pointSHP, "OID@", narray, "_ID")

# Create a temporary table from output3pointSHP dataset.
tempArray = arcpy.da.FeatureClassToNumPyArray(output3pointSHP, ('OID@',
                                                                'POINT_X',
                                                                'POINT_Y',
                                                                'POINT_Z'))

# Local counters
j = 0
surfID = 1

# Temporary lists
colB = []
colF = []
colG = []
colH = []
colP = []
colQ = []
colR = []

# Slice the table to set every three subsequent points apart
for i in range (1, nRows, 3):
	a = tempArray[i-1:i]
	b = tempArray[i:i+1]
	c = tempArray[i+1:i+2]

    # Fill the colB temporary list with increasing integer value for each three points
	for k in range (0, 3):
		colB.append(surfID)
	surfID += 1

	# Calculate orientation of the plane defined by three given points (Hasbargen, 2012)
	U1i = (a['POINT_Y'] - b['POINT_Y']) * (c['POINT_Z'] - b['POINT_Z']) - (c['POINT_Y'] - b['POINT_Y']) * (a['POINT_Z'] - b['POINT_Z'])
	for k in range (0, 3):
		colF.append(U1i[0]) #attribute: U1i
	U2j = -((a['POINT_X'] - b['POINT_X']) * (c['POINT_Z'] - b['POINT_Z']) - (c['POINT_X'] - b['POINT_X']) * (a['POINT_Z'] - b['POINT_Z']))
	for k in range (0, 3):
		colG.append(U2j[0]) #attribute: -U2j
	U3k = (a['POINT_X'] - b['POINT_X']) * (c['POINT_Y'] - b['POINT_Y']) - (c['POINT_X'] - b['POINT_X']) * (a['POINT_Y'] - b['POINT_Y'])
	for k in range (0, 3):
		colH.append(U3k[0]) #attribute: U3k

    # If three mentioned above points are vertexes of a triangle, calculate the position of each triangle centroid (barycenter)
	easting = (a['POINT_X'] + b['POINT_X'] + c['POINT_X']) / 3
	for k in range (0, 3):
		colP.append(easting[0]) #triangle centroid, attribute: easting
	northing = (a['POINT_Y'] + b['POINT_Y'] + c['POINT_Y']) / 3
	for k in range (0, 3):
		colQ.append(northing[0]) #triangle centroid, attribute: northing
	elevation = (a['POINT_Z'] + b['POINT_Z'] + c['POINT_Z']) / 3
	for k in range (0, 3):
		colR.append(elevation[0]) #triangle centroid, attribute: elevation

# Define fields for UpdateCoursor
fields = ['surfID', 'U1i', 'minusU2j', 'U3k', 'eastingFl', 'northingFl', 'strikeFl', 'dipFl', 'strike', 'dipDir', 'dip', 'easting', 'northing', 'elevation']

# Fill the surfID column with increasing attribute for each three points from temporary list and the rest of columns with results of above calculations
with arcpy.da.UpdateCursor(output3pointSHP, fields) as cursor:
	for row in cursor:
		row[0] = colB[j]
		row[1] = colF[j]
		row[2] = colG[j]
		row[3] = colH[j]
		if colH[j] < 0:
			row[4] = colG[j]
		elif colH[j] > 0:
			row[4] = -colG[j]
		if colH[j] > 0:
			row[5] = colF[j]
		elif colH[j] < 0:
			row[5] = -colF[j]
		if row[4] >= 0:
			row[6] = math.degrees(math.acos(row[5] / (row[4] ** 2 + row[5] ** 2) ** 0.5))
		elif row[4] < 0:
			row[6] = math.degrees(2 * math.pi - math.acos(row[5] / (row[4] ** 2 + row[5] ** 2) ** 0.5))
		row[7] = math.degrees(math.asin((colF[j] ** 2 + colG[j] ** 2) ** 0.5 / (colF[j] ** 2 + colG[j] ** 2 + colH[j] ** 2) ** 0.5))
		row[8] = int(round(row[6]))
		row[9] = row[8] + 90
		row[10] = int(round(row[7]))
		row[11] = colP[j]
		row[12] = colQ[j]
		row[13] = colR[j]
		cursor.updateRow(row)
		j += 1

# Create temporary layer with centroids
arcpy.MakeXYEventLayer_management(output3pointSHP, "easting", "northing", "centroidLayer", spatialReference, "elevation")

# Create dataset with centroids
arcpy.CopyFeatures_management("centroidLayer", outputCentroidSHP)
